module my_analysis
use my_own_da
implicit none
private concat,concatda,real_in_map,map_in_real,concatreal,map_zero,mat_map,map_mat
private unaryminus,POWMAP,POWMAPda,add_r_m,add_m_r,normalise,der,input_map_in_vec,input_vec_in_map,submap,subtaylor
private  comp_in_map,map_in_comp,print_my_vector_field,vector_field_zero,gradt,concatintaylor,texpt,input_map_in_my_taylor
private inverse_dragt_finn_zero,input_idf_map,CANONIZE_map

real(dp) :: epsresonance = 1.d-10 

 TYPE my_map
       type(my_taylor) v(2)   !@1 &nbsp; good old Taylor map of optics code
 end TYPE my_map
 
type(my_map) to_phasor,to_phasor_inv !@1 &nbsp; x+ip x-ip basis transformation
 
 TYPE vector_field         !@1 &nbsp;  map= exp(f.grad)I  
       integer n          !@1 &nbsp;  map= exp(f.grad)I expanded up to order f^n  
       type(my_taylor)v(2)    
 end TYPE vector_field
 
 TYPE inverse_dragt_finn
       type(my_map) m    
       type(vector_field) f    
 end TYPE inverse_dragt_finn
 

 TYPE normalform
       type(my_map) a_t   !@1 &nbsp; Canonical transformation such that Map= at r at^-1
       type(my_map) r    !@1 &nbsp; normal form : rotation usually
       real(dp) tune,damping   !@1 &nbsp; angle of rotation and damping if nonsymplectic
       real(dp) dtune_dA     !@1 &nbsp; Tune shift with amplitude A=x^2+p^2=2J 
 end TYPE normalform


  INTERFACE assignment (=)
     MODULE PROCEDURE input_my_taylor_in_map   !@1 &nbsp; map=Taylors
     MODULE PROCEDURE input_map_in_my_taylor   !@1 &nbsp; Taylors=map
     MODULE PROCEDURE input_my_map_in_map   !@1 &nbsp; map=map
     MODULE PROCEDURE input_vec_in_map   !@1 &nbsp; map=vector_field
     MODULE PROCEDURE input_map_in_vec   !@1 &nbsp; vector_field=map
     MODULE PROCEDURE input_idf_map   !@1 &nbsp; inverse dragt finn=map
     MODULE PROCEDURE real_in_map   !@1 &nbsp; map=real
     MODULE PROCEDURE map_in_real   !@1 &nbsp; real=map
     MODULE PROCEDURE comp_in_map  !@1 &nbsp; map=complex
     MODULE PROCEDURE map_in_comp  !@1 &nbsp; complex=map
     MODULE PROCEDURE map_zero   !@1 &nbsp; map=0  or map=identity  (syntax map=0 or map=1)
     MODULE PROCEDURE inverse_dragt_finn_zero !@1 &nbsp; idf=0  or idf=identity  (syntax idf=0 or idf=1)
     MODULE PROCEDURE vector_field_zero   !@1 &nbsp; vector_field=0     
     MODULE PROCEDURE mat_map   !@1 &nbsp;  mat=map
     MODULE PROCEDURE map_mat   !@1 &nbsp;  map=mat
     MODULE PROCEDURE mat_mapc   !@1 &nbsp;  mat=map  : complex matrix
     MODULE PROCEDURE map_matc   !@1 &nbsp;  map=mat  : complex matrix
     MODULE PROCEDURE normalise  !@1 &nbsp;  The normal form tools central to accelerator theory
     
  end  INTERFACE  
  
    INTERFACE OPERATOR (.o.)
!@3  <b><font color="#FF0000">map=map.o.map is interesting </font></b> click below </br>
     MODULE PROCEDURE concat   !@1 &nbsp; map 0 map   
     MODULE PROCEDURE concatreal   !@1 &nbsp; map 0 real(2)   
  END INTERFACE 
  
      INTERFACE OPERATOR (**) !@2 <font color="#FF0000"><b>Constant part is ignored in a &quot;DA&quot; calculation</b></font>
        MODULE PROCEDURE  POWMAPDA  !@1 &nbsp; map is raised to a power; can be negative (inverse)
      END INTERFACE 
      INTERFACE OPERATOR (.oo.) !@2 <font color="#FF0000"><b>Constant part is used in a &quot;TPSA&quot; calculation</b></font>
        MODULE PROCEDURE  POWMAP  !@1 &nbsp; map is raised to a power; can be negative (inverse)
      END INTERFACE 

    INTERFACE OPERATOR (*)
!@3  <b><font color="#FF0000">map=map*map is interesting </font></b> click below </br>
     MODULE PROCEDURE concatda   !@1 &nbsp; map * map  
     MODULE PROCEDURE concatintaylor   !@1 &nbsp; taylor o map
     MODULE PROCEDURE concatrealda   !@1 &nbsp; map 0 real(2) 
  END INTERFACE

    INTERFACE OPERATOR (+)
     MODULE PROCEDURE add_r_m   !@1 &nbsp;  ray+map  
     MODULE PROCEDURE add_m_r   !@1 &nbsp;  map+ray  
  END INTERFACE
  
  INTERFACE OPERATOR (.d.)
     MODULE PROCEDURE der   !@1 &nbsp;  derivative  
  END INTERFACE

  INTERFACE OPERATOR (.grad.)
     MODULE PROCEDURE gradt   !@1 &nbsp;  <font color="#FF0000"><b>F<font face="Symbol">~</font>M&nbsp; where F is type vector_field and M is type my_map</b></font>
  END INTERFACE

  INTERFACE OPERATOR (.subpart.)
     MODULE PROCEDURE subtaylor   !@1 &nbsp;  gets the ith part of the taylor  
     MODULE PROCEDURE submap   !@1 &nbsp;  gets the ith part of the map  
  END INTERFACE

  INTERFACE OPERATOR (.vf.)
     MODULE PROCEDURE extract_leading_vecfield  !@1 &nbsp;  gets the ith part of the map and put it in a vector field
  END INTERFACE

  INTERFACE OPERATOR (-)
     MODULE PROCEDURE unaryminus   !@1 &nbsp;  -map  
  END INTERFACE

  INTERFACE texp
     MODULE PROCEDURE texpt   !@1 &nbsp; <font color="#FF0000"><b>exp(F<font face="Symbol">~</font>)M where F is type vector_field and M is type my_map</b></font></font>  
  END INTERFACE

  INTERFACE exp
     MODULE PROCEDURE texpt !@1 &nbsp; <font color="#FF0000"><b>exp(F<font face="Symbol">~</font>)M where F is type vector_field and M is type my_map</b></font></font>  
  END INTERFACE

  INTERFACE CANONIZE
     MODULE PROCEDURE CANONIZE_map   !@1 &nbsp; Puts the canonical transformation into a special form
  END INTERFACE


   
  INTERFACE print
     MODULE PROCEDURE print_my_map   !@1 &nbsp; prints map for human eye
     module procedure print_my_vector_field !@1 &nbsp; prints a vector field for human eye
  END INTERFACE

contains


  subroutine map_mat( S2, S1 )
    implicit none
    real(dp), INTENT (IN) :: S1(2,2)
    TYPE (my_map), INTENT (inout) :: S2
    integer i,j 
      s2=0
      
      do i=1,2
      do j=1,2
       S2%v(i)%a(j)=s1(i,j)
      enddo
      enddo

  END subroutine map_mat
  
  subroutine mat_map( S2, S1 )
    implicit none
    real(dp), INTENT (INout) :: S2(2,2)
    TYPE (my_map), INTENT (in) :: S1
    integer i,j 
      
      s2=0.d0
      
      do i=1,2
      do j=1,2
       s2(i,j)=S1%v(i)%a(j)
      enddo
      enddo

  END subroutine mat_map
  
  subroutine map_matc( S2, S1 )
    implicit none
    complex(dp), INTENT (IN) :: S1(2,2)
    TYPE (my_map), INTENT (inout) :: S2
    integer i,j 
      s2=0
      
      do i=1,2
      do j=1,2
       S2%v(i)%a(j)=s1(i,j)
      enddo
      enddo

  END subroutine map_matc
  
  subroutine mat_mapc( S2, S1 )
    implicit none
    complex(dp), INTENT (INout) :: S2(2,2)
    TYPE (my_map), INTENT (in) :: S1
    integer i,j 
      
      s2=0.d0
      
      do i=1,2
      do j=1,2
       s2(i,j)=S1%v(i)%a(j)
      enddo
      enddo

  END subroutine mat_mapc


  function texpt(s2, S1 )
    implicit none
    TYPE (my_map) texpt
    TYPE (my_map),INTENT (IN) :: S1
    TYPE (vector_field),INTENT (IN) :: S2
    TYPE (my_map) t,sf
    integer i,j
    
    t=s1
    sf=s1
    
    do i=1,s2%n
     t=s2.grad.t
     do j=1,2
      t%v(j)=t%v(j)/REAL(i,kind=DP)   ! creating the exponential
      sf%v(j)=sf%v(j)+t%v(j)
     enddo
    enddo
    
    texpt=sf

  END function texpt

  function gradt(s2, S1 )
    implicit none
    TYPE (my_map) gradt
    TYPE (my_map),INTENT (IN) :: S1
    TYPE (vector_field),INTENT (IN) :: S2
    integer i,j
    TYPE (my_map) t
    
    t=0
    
    do j=1,2
    do i=1,2
     t%v(j)= s2%v(i)*(s1%v(j).d.i) + t%v(j)
    enddo
    enddo
    
    gradt=t
    
  END function gradt


  
  function der(s2, S1 )
    implicit none
    TYPE (my_taylor) der
    integer, INTENT (IN) :: S1
    TYPE (my_taylor),INTENT (IN) :: S2
    integer i,k

     if(first_mul) then
      call multiplication_table
      first_mul=.false.
     endif

      der=0.0_dp 
      
      if(s1==1) then
       do i=1,n_mono
        k=der_table(i,1)
        if(k>=0.and.jorder(k)<=my_order) then
         der%a(k)=jexp1(i)*s2%a(i)
        endif
       enddo
      endif

      if(s1==2) then
       do i=1,n_mono
        k=der_table(i,2)
        if(k>=0.and.jorder(k)<=my_order) then
         der%a(k)=jexp2(i)*s2%a(i)
        endif
       enddo
      endif

  END function der

  function subtaylor(s2, S1 )
    implicit none
    TYPE (my_taylor) subtaylor
    integer, INTENT (IN) :: S1
    TYPE (my_taylor),INTENT (IN) :: S2
    
      subtaylor=0.0_dp 
      if(s1==0) then
       subtaylor%a(0)=s2%a(0)
      endif
      if(s1==1) then
       subtaylor%a(1:2)=s2%a(1:2)
      endif
      if(s1==2) then
       subtaylor%a(3:5)=s2%a(3:5)
      endif
      if(s1==3) then
       subtaylor%a(6:9)=s2%a(6:9)
      endif
      if(s1==4) then
       subtaylor%a(10:n_mono)=s2%a(10:n_mono)
      endif

  END function subtaylor

  function submap(s2, S1 )
    implicit none
    TYPE (my_map) submap
    integer, INTENT (IN) :: S1
    TYPE (my_map),INTENT (IN) :: S2
    
    submap%v(1)=s2%v(1).subpart.s1
    submap%v(2)=s2%v(2).subpart.s1
    
  END function submap

  function extract_leading_vecfield(s2, S1 )
    implicit none
    TYPE (vector_field) extract_leading_vecfield
    integer, INTENT (IN) :: S1
    TYPE (my_map),INTENT (IN) :: S2
   
    extract_leading_vecfield%v(1)=s2%v(1).subpart.s1
    extract_leading_vecfield%v(2)=s2%v(2).subpart.s1
    
  END function extract_leading_vecfield


  subroutine map_zero( S2, S1 )
    implicit none
    integer, INTENT (IN) :: S1
    TYPE (my_map), INTENT (inout) :: S2
     
      S2%v(1)=0.d0
      S2%v(2)=0.d0
      
      if(s1==1) then
       S2%v(1)%a(1)=1.d0
       S2%v(2)%a(2)=1.d0
      endif

  END subroutine map_zero
  
    subroutine inverse_dragt_finn_zero( S2, S1 )
    implicit none
    integer, INTENT (IN) :: S1
    TYPE (inverse_dragt_finn), INTENT (inout) :: S2
     
      s2%m=s1
      s2%f=s1

  END subroutine inverse_dragt_finn_zero

  subroutine vector_field_zero( S2, S1 )
    implicit none
    integer, INTENT (IN) :: S1
    TYPE (vector_field), INTENT (inout) :: S2
    integer power
      S2%v(1)=0.0_dp
      S2%v(2)=0.0_dp
      
      power=s1
      if(power<=my_order) power=my_order+1
      s2%n=power

  END subroutine vector_field_zero

  subroutine input_idf_map( S2, S1 )
    implicit none
    TYPE (my_map), INTENT (IN) :: S1
    TYPE (inverse_dragt_finn), INTENT (inout) :: S2
    TYPE (my_map) m,DF,m0
    integer i
    
    s2=0
    s2%m=s1.subpart.1
    
    m=s2%m**(-1)*s1
   m0=m
    
    do i=1,my_order
    
     DF=m.subpart.1
     s2%f%V(1)=-(M%V(1)-DF%V(1))+s2%f%V(1)
     s2%f%V(2)=-(M%V(2)-DF%V(2))+s2%f%V(2)
       m=exp(s2%f,m0)
    
    enddo
     
     s2%f%V(1)=-s2%f%V(1)
     s2%f%V(2)=-s2%f%V(2)
    

    
  END subroutine input_idf_map



  
  subroutine input_my_taylor_in_map( S2, S1 )
    implicit none
    TYPE (my_taylor), INTENT (IN) :: S1(2)
    TYPE (my_map), INTENT (inout) :: S2

     S2%v(1)=S1(1)
     S2%v(2)=S1(2)
     
  END subroutine input_my_taylor_in_map
  
  subroutine input_map_in_my_taylor( S2, S1 )
    implicit none
    TYPE (my_taylor), INTENT (inout) :: S2(2)
    TYPE (my_map), INTENT (in) :: S1

     S2(1)=S1%v(1)
     S2(2)=S1%v(2)
     
  END subroutine input_map_in_my_taylor
  
  subroutine input_my_map_in_map( S2, S1 )
    implicit none
    TYPE (my_map), INTENT (IN) :: S1 
    TYPE (my_map), INTENT (inout) :: S2

     S2%v(1)=S1%v(1)
     S2%v(2)=S1%v(2)
     
  END subroutine input_my_map_in_map

  subroutine input_vec_in_map( S2, S1 )
    implicit none
    TYPE (vector_field), INTENT (IN) :: S1 
    TYPE (my_map), INTENT (inout) :: S2

     S2%v(1)=S1%v(1)
     S2%v(2)=S1%v(2)
     
  END subroutine input_vec_in_map

  subroutine input_map_in_vec( S2, S1 )
    implicit none
    TYPE (my_map), INTENT (IN) :: S1 
    TYPE (vector_field), INTENT (inout) :: S2

     S2%v(1)=S1%v(1)
     S2%v(2)=S1%v(2)
     
  END subroutine input_map_in_vec

  
   subroutine real_in_map( S2, S1 )
    implicit none
    real(dp), INTENT (IN) :: S1(2)
    TYPE (my_map), INTENT (inout) :: S2

     S2%v(1)=S1(1)
     S2%v(2)=S1(2)
     
  END subroutine real_in_map
 
  subroutine map_in_real( S2, S1 )
    implicit none
    real(dp), INTENT (INOUT) :: S2(2)
    TYPE (my_map), INTENT (in) :: S1

     S2(1)=S1%v(1)
     S2(2)=S1%v(2)
     
  END subroutine map_in_real
 
    subroutine comp_in_map( S2, S1 )
    implicit none
    complex(dp), INTENT (IN) :: S1(2)
    TYPE (my_map), INTENT (inout) :: S2

     S2%v(1)=S1(1)
     S2%v(2)=S1(2)
     
  END subroutine comp_in_map

  subroutine map_in_comp( S2, S1 )
    implicit none
    complex(dp), INTENT (INOUT) :: S2(2)
    TYPE (my_map), INTENT (in) :: S1

     S2(1)=S1%v(1)
     S2(2)=S1%v(2)
     
  END subroutine map_in_comp
 
  subroutine print_my_map( S2, mf )
    implicit none
    TYPE (my_map), INTENT (inout) :: S2
    integer mf
    write(mf,*) " variable 1"
     call print(s2%v(1),mf)
     write(mf,*) " variable 2"
     call print(s2%v(2),mf)

  END subroutine print_my_map
  
  subroutine print_my_vector_field( S2, mf )
    implicit none
    TYPE (vector_field), INTENT (inout) :: S2
    integer mf
    write(mf,*) " variable 1"
     call print(s2%v(1),mf)
     write(mf,*) " variable 2"
     call print(s2%v(2),mf)

  END subroutine print_my_vector_field
  
  FUNCTION unaryminus( S2 )
    implicit none
     TYPE (my_map) unaryminus
     TYPE (my_map), INTENT (in) :: S2
     integer i
    
     unaryminus%v(1)= -S2%v(1)
     unaryminus%v(2)= -S2%v(2)
     
  END FUNCTION unaryminus
 

  FUNCTION add_r_m( S1, S2 )
    implicit none
    TYPE (my_taylor) add_r_m(2)
    real(dp), INTENT (IN) :: S1(2) 
    TYPE (my_map), INTENT (IN) :: s2

     add_r_m(1)=s2%v(1)+s1(1)
     add_r_m(2)=s2%v(2)+s1(2)
     
  end FUNCTION add_r_m

  FUNCTION add_m_r( S2, S1 )
    implicit none
    TYPE (my_taylor) add_m_r(2)
    real(dp), INTENT (IN) :: S1(2) 
    TYPE (my_map), INTENT (IN) :: s2

     add_m_r(1)=s2%v(1)+s1(1)
     add_m_r(2)=s2%v(2)+s1(2)
     
  end FUNCTION add_m_r

  FUNCTION concat( S1, S2 )
    implicit none
    TYPE (my_map) concat
    TYPE (my_map), INTENT (IN) :: S1, S2
    TYPE (my_taylor)  x1,x2,y(2),z(2)
    integer i
    x1=s2%v(1)
    x2=s2%v(2)
    y(1)=s1%v(1)
    y(2)=s1%v(2)
    
    z(1)=0.d0
    z(2)=0.d0
    
! " (0,0) ", S%A(0)
! " (1,0) ", S%A(1)
! " (0,1) ", S%A(2)
! " (2,0) ", S%A(3)
! " (0,2) ", S%A(4)
! " (1,1) ", S%A(5)
! " (3,0) ", S%A(6)
! " (0,3) ", S%A(7) 
! " (2,1) ", S%A(8)
! " (1,2) ", S%A(9)
!@3  <p><i><sup><font size="4" color="#0066FF">Lazy implementation: should be  generalized, only good to 4th order.</font></sup></i></p>

   do i=1,2
      z(i)=y(i)%a(0) 
  	 if(my_order>=1) z(i)=y(i)%a(1)*x1+y(i)%a(2)*x2+z(i)
     if(my_order>=2) z(i)=y(i)%a(3)*x1**2+y(i)%a(4)*x2**2+y(i)%a(5)*x1*x2+z(i)
     if(my_order>=3) z(i)=y(i)%a(6)*x1**3+y(i)%a(7)*x2**3+y(i)%a(8)*x1**2*x2+y(i)%a(9)*x1*x2**2+z(i)
     if(my_order>=4) z(i)=y(i)%a(10)*x1**4+y(i)%a(11)*x2**4+y(i)%a(12)*x1**3*x2+y(i)%a(13)*x1*x2**3+y(i)%a(14)*x1**2*x2**2+z(i)
    enddo
  
    concat%v(1)=z(1)
    concat%v(2)=z(2)
     
  END FUNCTION concat


  FUNCTION concatda( S1, S2 )
    implicit none
    TYPE (my_map) concatda
    TYPE (my_map), INTENT (IN) :: S1, S2
    TYPE (my_map)   S20
      
    s20=s2
    s20%v(1)%a(0)=0.d0
    s20%v(2)%a(0)=0.d0
 

      concatda=S1.O.S20

 
  END FUNCTION concatda

  FUNCTION concatintaylor( T, S2 )
    implicit none
    TYPE (my_taylor) concatintaylor
    TYPE (my_taylor), INTENT (IN) :: t
    TYPE (my_map), INTENT (IN) :: S2
    TYPE (my_map)   S20
      
    s20=0
    
    s20%v(1)=t
     
     s20=S20*S2

      concatintaylor=s20%v(1)

 
  END FUNCTION concatintaylor



  FUNCTION concatreal( S1, z )
    implicit none
    real(dp)  concatreal(2)
    TYPE (my_map), INTENT (IN) :: S1
    real(dp), INTENT (IN) :: z(2)
    TYPE (my_map) zm
    
    zm=z

    zm=s1.o.zm
 
    concatreal=zm
     
  END FUNCTION concatreal


  FUNCTION concatrealda( S1, z )
    implicit none
    real(dp)  concatrealda(2)
    TYPE (my_map), INTENT (IN) :: S1
    real(dp), INTENT (IN) :: z(2)
    TYPE (my_map) zm,s10
    
    s10=s1
    s10%v(1)%a(0)=0.d0
    s10%v(2)%a(0)=0.d0
    
    zm=z

    zm=s10.o.zm
 
    concatrealda=zm
     
  END FUNCTION concatrealda



  subroutine invert_map( S1, S2 )
    implicit none
    TYPE (my_map), INTENT (IN) :: S1 
    TYPE (my_map), INTENT (inout) :: S2
    TYPE (my_map) at,s10,a
    complex(dp) det,mi(2,2),d(2)
    integer i
    
    mi(1,1)=s1%v(2)%a(2)
    mi(2,2)=s1%v(1)%a(1)
    mi(1,2)=-s1%v(1)%a(2)
    mi(2,1)=-s1%v(2)%a(1)
    det=mi(1,1)*mi(2,2)-mi(1,2)*mi(2,1)
    mi=mi/det
    
    s10=s1
    s10%v(1)%a(0)=0.d0
    s10%v(2)%a(0)=0.d0
    
    at=0   
    at=mi
    a=at
    s10=at*s10
    do i=2, my_order
    at=s10
     at%v(1)%a(1:2)= -at%v(1)%a(1:2)
     at%v(2)%a(1:2)= -at%v(2)%a(1:2)
     at=-at
     s10=at*s10
     a=at*a    
    enddo
    d=S1
    at=1
    at%v(1)=at%v(1)-d(1)
    at%v(2)=at%v(2)-d(2)
    s2=a.o.at
  END subroutine invert_map

  FUNCTION POWMAP( S1, R2 )
    implicit none
    TYPE (my_map) POWMAP
    TYPE (my_map), INTENT (IN) :: S1
    INTEGER, INTENT (IN) :: R2
    TYPE (my_map) S11
    INTEGER I,R22

    s11=1


    R22=IABS(R2)
    DO I=1,R22
       s11=s1*s11
    ENDDO

    IF(R2.LT.0) THEN
       CALL invert_map( S11, S11 ) 
    ENDIF

    powmap=s11

  END FUNCTION POWMAP

  FUNCTION POWMAPda( S1, R2 )
    implicit none
    TYPE (my_map) POWMAPda
    TYPE (my_map), INTENT (IN) :: S1
    INTEGER, INTENT (IN) :: R2
    TYPE (my_map) S11,S10
    INTEGER I,R22

    s11=1
    s10=S1
    S10%V(1)%A(0)=0.D0
    S10%V(2)%A(0)=0.D0

    R22=IABS(R2)
    DO I=1,R22
       s11=s10*s11
    ENDDO

    IF(R2.LT.0) THEN
       CALL invert_map( S11, S11 ) 
    ENDIF
    POWMAPda=s11

  END FUNCTION POWMAPda

  subroutine CANONIZE_map( A,A_cs,R,PHASE_ADVANCE )
    implicit none
    TYPE (my_map) , INTENT (inout) ::   A,A_cs,R
    TYPE(MY_TAYLOR), INTENT (inout) ::  PHASE_ADVANCE
    TYPE (my_map) RT,AN
    type(inverse_dragt_finn) idf,idfn
    type(vector_field) f

    REAL(DP) C,S,D_MU
    
    R=1
    D_MU=ATAN2(real(A%V(1)%A(2),KIND=DP),real(A%V(1)%A(1),KIND=DP))
    C=COS(D_MU)
    S=SIN(D_MU)
    PHASE_ADVANCE=PHASE_ADVANCE+D_MU/TWOPI
    RT=0
    
    RT%V(1)%A(1)=C
    RT%V(1)%A(2)=S
    RT%V(2)%A(2)=C
    RT%V(2)%A(1)=-S
    
    A_cs=A*RT**(-1)
    
    R=RT*R
    
    
    call create_phasors
    
    idf=a_cs
    AN=1
    AN=exp(idf%f,AN)
    AN=to_phasor *AN*to_phasor_inv
    idfn=an
    
    f=0
    f%v(1)%a(8)=idfn%f%v(1)%a(8)
    f%v(2)%a(9)=idfn%f%v(2)%a(9)
    RT=1
    RT=exp(f,RT)
    RT=to_phasor_inv*RT*to_phasor

    R=RT*R
    A_CS=A_CS*RT**(-1)

    PHASE_ADVANCE=PHASE_ADVANCE-((1.0_DP.MONO.1)**2+(1.0_DP.MONO.2)**2)*IMAG(idfn%f%v(1)%a(8))/TWOPI
    
    
  END subroutine CANONIZE_map

  subroutine normalise( N, M )
    implicit none
    TYPE (normalform), INTENT (inout) ::N
    TYPE (my_map), INTENT (in) :: M
    real(dp) mat(2,2),det,c,s,ang,alpha,beta,dam
    TYPE (my_map) A1,R,NL,nlc
    type(vector_field) f
    integer i

    n%dtune_dA=0.0_dp
    call create_phasors

!@3 !!!!  Linear Analysis starts here

    mat=m
    
    det=(mat(1,1)*mat(2,2)-mat(1,2)*mat(2,1))  

    
    mat=mat/sqrt(det)
    c=(mat(1,1)+mat(2,2))/2.0_dp  
    
    ang=acos(c)         !@2 !!!!  getting tune
    
    if(mat(1,2)<0) ang=-ang+twopi
    n%tune=ang/twopi
    n%DAMPING=LOG(DET)/2.0_DP
    dam=n%DAMPING
    beta=mat(1,2)/sin(ang)
    alpha=(mat(1,1)-c)/sin(ang)
    
    a1=0
    a1%v(1)%a(1)=sqrt(beta)
    a1%v(2)%a(2)=1.0_dp/sqrt(beta)
    a1%v(2)%a(1)=-alpha/sqrt(beta)
    
    n%a_t=a1
 
 !@3 !!!! Linear normalisation done
 
    N%R=A1**(-1)*M*A1
    R=N%R.SUBPART.1  ! LINEAR PART
    NL=N%R*R**(-1)   ! NONLINEAR PART   N%R= NL * R 
    
    do i=2,my_order
     call analyse_kernel(i,ang,dam,nl,f,n%dtune_dA)
     a1=1    
     a1=exp(f,a1)
     a1=to_phasor_inv *a1*to_phasor
     
     n%r=a1**(-1)*n%r*a1
     NL=N%R*R**(-1)   ! NONLINEAR PART   N%R= NL * R 
     n%a_t=n%a_t*a1

    enddo
    
   END subroutine normalise
   
    
    
  subroutine analyse_kernel(order,ang,dam,nr,f,dtune_dA)
    implicit none
    TYPE (my_map), INTENT (inout) :: nr
    type(vector_field), INTENT (inout) ::  f
    TYPE (my_map) t
    integer i,j,k,order,i1,i2
    type(vector_field) g
    real(dp) dam,ang,dtune_dA
    complex(dp) lam(2),denominator

!@3  <p><i><sup><font size="4" color="#0066FF">Lazy implementation: should be  generalized, only good to 4th order.</font></sup></i></p>
   
    if(order==0) then
     i1=0
     i2=0
    elseif(order==1) then
     i1=1
     i2=2
    elseif(order==2) then
     i1=3
     i2=5
    elseif(order==3) then
     i1=6
     i2=9
    elseif(order==4) then
     i1=10
     i2=n_mono
    endif



    
    lam(1)=exp(-i_*ang+dam)
    lam(2)=exp(i_*ang+dam)
    
    f=0
    
   

     t=to_phasor*nr*to_phasor_inv 
     g=t.vf.order
     
        
      do i=1,2
      do j=i1,i2     
 
       denominator= 1.0_dp-lam(i)*exp(i_*(jexp1(j)-jexp2(j))*ang-(jexp1(j)+jexp2(j))*dam)
       
       if(epsresonance>0) then    !@1 &nbsp; determined by closeness to a resonance
           if(abs(denominator)>epsresonance) then
               f%v(i)%a(j) = g%v(i)%a(j)/denominator 
           else
            if(jexp1(j)-jexp2(j)-1==0.and.my_order>=3.and.i==1) dtune_dA= i_*g%v(i)%a(j)/twopi !@1 &nbsp; Real part is tune shift
           endif
       else             !@1 &nbsp; leaves tune shifts and amplitude dependent damping
           if((jexp1(j)-jexp2(j)-1/=0.and.i==1).or.(jexp1(j)-jexp2(j)+1/=0.and.i==2)) then
               f%v(i)%a(j) = g%v(i)%a(j)/denominator 
           else
            if(my_order>=3.and.i==1) then
              dtune_dA= i_*g%v(i)%a(j)/twopi     !@1 &nbsp; Real part is tune shift, complex part  is damping (nonlinear)
             endif 
           endif
       endif      
      enddo
      enddo
   
    
      


  END subroutine analyse_kernel
 
   subroutine create_phasors(use_J) 
    implicit none
    logical, optional :: use_J
    real(dp) x
    
    x=1.0_dp
    if(present(use_J)) then
     if(use_J) x=sqrt(2.0_dp)
    endif
     !@1 &nbsp; complex transformation to phasors basis
     
    to_phasor=0
    to_phasor%v(1)%a(1)=1.0_dp/x
    to_phasor%v(1)%a(2)=i_/x
    to_phasor%v(2)%a(1)=1.0_dp/x
    to_phasor%v(2)%a(2)=-i_/x
    
    to_phasor_inv=to_phasor**(-1)
   end subroutine create_phasors 

!!! Some useful routines

   subroutine AVERAGE(F,A,F_floquet,F_xp,use_J) 
    implicit none
    type(my_map) A
    TYPE(MY_TAYLOR) F,F_floquet
    TYPE(MY_TAYLOR), optional :: F_xp
    logical, optional :: use_J
    logical usej
    integer i

    !! USE_J is false by default
         
    CALL create_phasors(use_J)
    
    F_floquet=(F*A)*to_phasor**(-1)
    
    do i=0,n_mono
      if(jorder(i)>my_order) cycle
      if(jexp1(i)/=jexp2(i)) then 
       F_floquet%a(i)=0.0_dp
      endif       
    enddo
    
    if(present(F_xp)) F_xp=F_floquet*to_phasor*A**(-1)
    

   end subroutine AVERAGE 


end module my_analysis